<?php

declare(strict_types = 1);

namespace Drupal\musica\Form;

use Drupal\musica\Form\SoundCloudConsent;
use Drupal\Core\Form\FormStateInterface;

/**
 * Authorization form for SoundCloud service.
 *
 * @see https://www.drupal.org/docs/drupal-apis/services-and-dependency-injection/dependency-injection-for-a-form
 * @see https://www.drupal.org/docs/develop/creating-modules/defining-and-using-your-own-configuration-in-drupal
 */
final class SoundCloudAuth extends SoundCloudConsent {

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'musica_soundcloud_auth';
  }

  private function currentAccessToken($type = 'access'): array {
    $element = [];

    if ($this->service->isAuthorized()) {
      $token = $this->service->getTokenObject($type);
      $element = [
        '#type' => 'textarea',
        '#title' => $this->t("Your current $type token."),
        '#rows' => 3,
        '#resizable' => 'none',
        '#value' => var_export($token, TRUE),
        '#disabled' => TRUE,
      ];
    }

    return $element;
  }

  /**
   * This Form authorizes the application to use the Spotify API.
   *
   * @see https://packagist.org/packages/martin1982/oauth2-soundcloud
   * @see https://api.drupal.org/api/drupal/elements/10
   * @see https://api.drupal.org/api/drupal/core%21lib%21Drupal%21Core%21Render%21Element%21FormElement.php/class/FormElement/10
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $authorized = $this->service->isAuthorized();

    if (array_key_exists('code', $_GET) && !empty($_GET['code'])) {
      $authorized = $this->service->getAccessToken('authorization_code', ['code' => $_GET['code']]);

      // If we got here through an outdated redirect URL, make sure we send the user to the correct location gracefully.
      $this->messenger()->addStatus($this->t("Your application has been authorized. <br />You will be redirected to Musica's configuration page in 10 seconds."));
      header('Refresh: 10; URL=/admin/config/musica/soundcloud/authorize');
    }

    $form['about'] = [
      '#type' => 'item',
      '#title' => $this->t('SoundCloud® API'),
      '#description' => $this->t('For more information about the SoundCloud API visit the <a href="https://developers.soundcloud.com/">SoundCloud Developer Portal</a>'),
    ];

    $form['description'] = [
      '#type' => 'item',
      '#title' => $this->t('Authorization'),
      '#description' => $this->t('This page allows your Drupal application to connect with the SoundCloud API using the parameters you entered in the <a href="/admin/config/musica/soundcloud/configure">Configure</a> page.'),
    ];

    $text_not_authorized = $this->t('Your application is currently not authorized. </br>
      To authorize your application click on the "Authorize" button. </br>
      <b>Notice: You will be automatically re-directed to SoundCloud for authorization.</b>');
    $text_authorized = $this->t('Your application is currently authorized.');

    $form['authorized'] = [
      '#type' => 'item',
      '#title' => $this->t('Status'),
      '#description' => $authorized ? $text_authorized : $text_not_authorized,
      '#return_value' => $authorized,
      '#default_value' => $authorized,
      '#disabled' => !$authorized,
    ];

    $form['access_token'] = $this->currentAccessToken('access');
    $form['refresh_token'] = $this->currentAccessToken('refresh');

    $form['actions'] = ['#type' => 'actions'];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Authorize'),
      '#button_type' => 'primary',
      '#disabled' => $authorized,
    ];
    $form['actions']['revoke'] = [
      '#type' => 'button',
      '#value' => $this->t('Revoke'),
      '#disabled' => !$authorized,
    ];
    return $form;
  }

  public function validateForm(array &$form, FormStateInterface $form_state) {

    // If the form is NOT "submitted" but it is being validated, it means the
    // revoke action has been invoked.
    if (!$form_state->isSubmitted()) {
      $this->service->revoke();
      $this->messenger()->addWarning($this->t("Musica has revoked your application's SoundCloud authorization.</br>
        You may still request re-authorization, if so desired."));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {

    if (!isset($_GET['code'])) {
      // If we don't have an authorization code then get one.
      $authUrl = $this->service->provider->getAuthorizationUrl();
      $_SESSION['oauth2state'] = $this->service->provider->getState();
      header('Location: ' . $authUrl);
      exit;
    }
    // Check given state against previously stored one to mitigate CSRF attack
    elseif (empty($_GET['state']) || ($_GET['state'] !== $_SESSION['oauth2state'])) {
      unset($_SESSION['oauth2state']);
      echo 'Invalid OAuth authorization state.';
      exit;
    }

  }

}
