<?php

declare(strict_types = 1);

namespace Drupal\musica\Form;

use Drupal\musica\Service\SoundCloud;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\State\State;
use Drupal\Core\Messenger\MessengerTrait;

/**
 * Authorization form for SoundCloud service.
 *
 * @see https://www.drupal.org/docs/drupal-apis/services-and-dependency-injection/dependency-injection-for-a-form
 * @see https://www.drupal.org/docs/develop/creating-modules/defining-and-using-your-own-configuration-in-drupal
 */
class SoundCloudConsent extends FormBase {
  use MessengerTrait;

  /**
   * State API item id for `terms` form element.
   */
  const TOS_CONSENT = 'musica:soundcloud:tos_consent';

  protected State $state;
  protected SoundCloud $service;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'musica_auth_soundcloud';
  }

  /**
   * {@inheritdoc}
   */
  public function __construct(State $state, SoundCloud $service) {
    $this->state = $state;
    $this->service = $service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('state'),
      $container->get('musica.soundcloud')
    );
  }

  /**
   * Whether the user has consented to the SoundCloud Terms of Use or not.
   */
  protected function hasConsent(): bool {
    return $this->state->get(self::TOS_CONSENT) !== NULL ? $this->state->get(self::TOS_CONSENT) : FALSE;
  }

  protected function consentElement(): array {
    if ($this->hasConsent()) {
      return [];
    }
    else {
      $form['terms'] = [
        '#type' => 'item',
        '#title' => $this->t('SoundCloud® Terms of Use'),
        '#description' => $this->t('<b>NOTICE</b>: In order to use the SoundCloud API you must first consent to their <a href="/admin/config/musica/soundcloud/consent">Terms of Use</a>'),
        '#required' => TRUE,
      ];
      return $form;
    }
  }

  /**
   * {@inheritdoc}
   *
   * @see https://api.drupal.org/api/drupal/elements/10
   * @see https://api.drupal.org/api/drupal/core%21lib%21Drupal%21Core%21Render%21Element%21FormElement.php/class/FormElement/10
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $form['description'] = [
      '#type' => 'item',
      '#title' => $this->t('SoundCloud® Terms of Use'),
      '#description' => $this->t('For more information about SoundCloud\'s Terms of Use visit  <a href="https://developers.soundcloud.com/docs/api/terms-of-use">https://developers.soundcloud.com/docs/api/terms-of-use</a>'),
    ];

    // If the user already consented to the TOS, gray out the element.
    $consented = $this->hasConsent();
    $consent = $form_state->hasValue('terms') ? $form_state->get('terms') : FALSE;
    $form['terms'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('I have fully read and agree to follow SoundCloud\'s <a href="https://developers.soundcloud.com/docs/api/terms-of-use">API Terms of Use</a>.'),
      '#description' => $this->t('<b>NOTICE</b>: You will not be allowed to configure and use the application until you have read and agreed to the Terms of Use linked above.'),
      '#return_value' => TRUE,
      '#required' => TRUE,
      '#default_value' => $consented ? $consented : $consent,
      '#disabled' => $consented ? $consented : $consent,
    ];

    $form['actions'] = [
      '#type' => 'actions',
    ];
    $form['actions']['#type'] = 'actions';
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Consent'),
      '#button_type' => 'primary',
      '#disabled' => $consented ? $consented : $consent,
    ];
    return $form;
  }

  public function validateForm(array &$form, FormStateInterface $form_state) {

    // The #required property requires consent to proceed, this is a backup.
    if ($form_state->getValue('terms') !== TRUE) {
      $form_state->setErrorByName('auth_confirmation', $this->t('YOU MUST READ AND AGREE TO SOUNDCLOUD\'s "API TERMS OF USE" BEFORE PROCEEDING.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {

    $this->state->set(self::TOS_CONSENT, $form_state->getValue('terms'));
  }

}
